import numpy as np
from numpy.testing import assert_almost_equal
from scipy.stats import kurtosis, skew

from preliz.distributions import DiscreteWeibull


def test_disc_weibull_vs_random():
    dist = DiscreteWeibull(0.7, 0.9)

    rng = np.random.default_rng(1)
    rvs = dist.rvs(500_000, random_state=rng)

    assert_almost_equal(dist.mean(), rvs.mean(), decimal=2)
    assert_almost_equal(dist.median(), np.median(rvs), decimal=2)
    assert_almost_equal(dist.var(), rvs.var(), decimal=1)
    assert_almost_equal(dist.std(), rvs.std(), decimal=1)
    assert_almost_equal(dist.skewness(), skew(rvs), decimal=1)
    assert_almost_equal(dist.kurtosis(), kurtosis(rvs), decimal=0)

    q = np.linspace(0.1, 0.9, 10)
    assert_almost_equal(dist.ppf(q), np.quantile(rvs, q), decimal=2)

    for x in np.arange(6):
        assert_almost_equal(dist.pdf(x), np.mean(rvs == x), decimal=2)
        assert_almost_equal(dist.cdf(x), np.sum(dist.pdf(np.arange(x + 1))), decimal=2)
        assert_almost_equal(dist.cdf(x), np.sum(np.exp(dist.logpdf(np.arange(x + 1)))), decimal=2)
        assert_almost_equal(dist.cdf(x), np.mean(rvs <= x), decimal=2)



"""Functions to communicate with Bambi."""

import importlib
import inspect
import re
from copy import copy
from sys import modules

import numpy as np


def get_pymc_model(model):
    if not model.built:
        model.build()
    pymc_model = model.backend.model
    return pymc_model


def write_bambi_string(new_priors, var_info):
    """
    Return a string with the new priors for the Bambi model.

    So the user can copy and paste, ideally with none to minimal changes.
    """
    header = "{\n"
    variables = []
    names = list(new_priors.keys())
    for key, value in new_priors.items():
        idxs = var_info[key][-1]
        if idxs:
            for i in idxs:
                nkey = names[i]
                cp_dist = copy(new_priors[nkey])
                cp_dist._fit_moments(np.mean(value.mean()), np.mean(value.std()))

                dist_name, dist_params = repr(cp_dist).split("(")
                size = var_info[nkey][1]
                if size > 1:
                    variables[
                        i
                    ] = f'"{nkey}" : bmb.Prior("{dist_name}", {dist_params}, shape={size}),\n'
                else:
                    variables[i] = f'"{nkey}" : bmb.Prior("{dist_name}", {dist_params}),\n'
        else:
            dist_name, dist_params = repr(value).split("(")
            dist_params = dist_params.rstrip(")")
            size = var_info[key][1]
            if size > 1:
                variables.append(
                    f'"{key}" : bmb.Prior("{dist_name}", {dist_params}, shape={size}),\n'
                )
            else:
                variables.append(f'"{key}" : bmb.Prior("{dist_name}", {dist_params}),\n')

    return "".join([header] + variables)


def from_bambi(fmodel, draws):
    module_name = fmodel.__module__
    module = importlib.import_module(module_name)

    # Get the source code of the original function
    original_source = inspect.getsource(fmodel)

    # Define a pattern to find the line where the model is built
    pattern = re.compile(r"(\s+)([a-zA-Z_]\w*)\s*=\s*.*?Model(.*)")

    # Find the match in the source code
    match = pattern.search(original_source)

    # Extract the indentation and variable name
    indentation = match.group(1)
    variable_name = match.group(2)

    # Find the variables after the return statement
    return_variables = match_return_variables(original_source)

    if return_variables:
        # Build the new source code
        new_source = original_source.replace(
            match.group(0),
            f"{match.group(0)}"
            f"{indentation}{variable_name}.build()\n"
            f"{indentation}variables = [{variable_name}.backend.model.named_vars[v] "
            f"for v in {return_variables}]\n"
            f'{indentation}{", ".join(return_variables)} = pm.draw(variables, draws={draws})',
        )

        # Find the priors we want to change
        all_dist_str = dist_as_str()
        matches = match_preliz_dist(all_dist_str, new_source)
        # Create a dictionary with the priors
        model = dict_model(matches, return_variables)

        # Execute the new source code to redefine the function
        exec(new_source, module.__dict__)
        modified_fmodel = getattr(module, fmodel.__name__)

    return modified_fmodel, return_variables, model


def match_preliz_dist(all_dist_str, source):
    # remove comments
    source = re.sub(r"#.*$|^#.*$", "", source, flags=re.MULTILINE)
    regex = rf'\s*(?:\w+\.)?Prior\("({all_dist_str})",\s*((?:\w+=\w+(?:,?\s*)?)*)\s*\)'
    matches = re.finditer(regex, source)
    return matches


def match_return_variables(source):
    match = re.search(r"return (\w+(\s*,\s*\w+)*)", source)
    return [var.strip() for var in match.group(1).split(",")]


def dist_as_str():
    all_distributions = modules["preliz.distributions"].__all__
    return "|".join(all_distributions)


def dict_model(matches, return_variables):
    model = {}
    obs_rv = return_variables[-1]
    for match in matches:
        var_name = match.group(0).split("=")[0].strip()
        if var_name != obs_rv:
            dist = getattr(modules["preliz.distributions"], match.group(2))
            model[var_name] = dist()

    return model



import numba as nb
import numpy as np

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml


class AsymmetricLaplace(Continuous):
    r"""
    Asymmetric-Laplace distribution.

    The pdf of this distribution is

    .. math::
        {f(x|\\b,\kappa,\mu) =
            \left({\frac{\\b}{\kappa + 1/\kappa}}\right)\,e^{-(x-\mu)\\b\,s\kappa ^{s}}}

    where

    .. math::

        s = sgn(x-\mu)

    .. plot::
        :context: close-figs


        from preliz import AsymmetricLaplace, style
        style.use('preliz-doc')
        kappas = [1., 2., .5]
        mus = [0., 0., 3.]
        bs = [1., 1., 1.]
        for kappa, mu, b in zip(kappas, mus, bs):
            AsymmetricLaplace(kappa, mu, b).plot_pdf(support=(-10,10))

    ========  ========================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu-\frac{\\\kappa-1/\kappa}b`
    Variance  :math:`\frac{1+\kappa^{4}}{b^2\kappa^2 }`
    ========  ========================

    AsymmetricLaplace distribution has 2 alternative parametrizations. In terms of kappa,
    mu and b or q, mu and b.

    The link between the 2 alternatives is given by

    .. math::

       \kappa = \sqrt(\frac{q}{1-q})

    Parameters
    ----------
    kappa : float
        Symmetry parameter (kappa > 0).
    mu : float
        Location parameter.
    b : float
        Scale parameter (b > 0).
    q : float
        Symmetry parameter (0 < q < 1).
    """

    def __init__(self, kappa=None, mu=None, b=None, q=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(kappa, mu, b, q)

    def _parametrization(self, kappa=None, mu=None, b=None, q=None):
        if all_not_none(kappa, q):
            raise ValueError("Incompatible parametrization. Either use kappa or q.")

        self.param_names = ("kappa", "mu", "b")
        self.params_support = ((eps, np.inf), (-np.inf, np.inf), (eps, np.inf))

        if q is not None:
            self.q = q
            kappa = self._from_q(q)
            self.param_names = ("q", "mu", "b")
            self.params_support = ((eps, 1 - eps), (-np.inf, np.inf), (eps, np.inf))

        self.kappa = kappa
        self.mu = mu
        self.b = b
        if all_not_none(kappa, mu, b):
            self._update(kappa, mu, b)

    def _from_q(self, q):
        kappa = (q / (1 - q)) ** 0.5
        return kappa

    def _to_q(self, kappa):
        q = kappa**2 / (1 + kappa**2)
        return q

    def _update(self, kappa, mu, b):
        self.kappa = np.float64(kappa)
        self.mu = np.float64(mu)
        self.b = np.float64(b)
        self.q = self._to_q(self.kappa)

        if self.param_names[0] == "kappa":
            self.params = (self.kappa, self.mu, self.b)
        elif self.param_names[0] == "q":
            self.params = (self.q, self.mu, self.b)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.b, self.kappa))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.b, self.kappa)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.b, self.kappa)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.b, self.kappa)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.b, self.kappa)

    def entropy(self):
        return nb_entropy(self.b, self.kappa)

    def median(self):
        if self.kappa > 1:
            return self.mu + self.kappa * self.b * np.log(
                (1 + self.kappa**2) / (2 * self.kappa**2)
            )
        return self.mu - np.log((1 + self.kappa**2) / 2) / (self.kappa / self.b)

    def mean(self):
        return (1 / self.kappa - self.kappa) * self.b + self.mu

    def mode(self):
        return self.mu

    def var(self):
        return ((1 / self.kappa) ** 2 + self.kappa**2) * self.b**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 2.0 * (1 - np.power(self.kappa, 6)) / np.power(1 + np.power(self.kappa, 4), 1.5)

    def kurtosis(self):
        return 6.0 * (1 + np.power(self.kappa, 8)) / np.power(1 + np.power(self.kappa, 4), 2)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(-self.kappa, 1 / self.kappa, size)
        return nb_rvs(random_samples, self.mu, self.b, self.kappa)

    def _fit_moments(self, mean, sigma):
        # Assume symmetry
        mu = mean
        b = (sigma / 2) * (2**0.5)
        self._update(1, mu, b)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, mu, b, kappa):
    x = (x - mu) / b
    kap_inv = 1 / kappa
    kap_kapinv = kappa + kap_inv
    if x >= 0:
        return 1 - np.exp(-x * kappa) * (kap_inv / kap_kapinv)
    return np.exp(x * kap_inv) * (kappa / kap_kapinv)


@nb.vectorize(nopython=True, cache=True)
def nb_ppf(q, mu, b, kappa):
    kap_inv = 1 / kappa
    kap_kapinv = kappa + kap_inv
    if q >= kappa / kap_kapinv:
        q_ppf = -np.log((1 - q) * kap_kapinv * kappa) * kap_inv
    else:
        q_ppf = np.log(q * kap_kapinv / kappa) * kappa
    return q_ppf * b + mu


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, b, kappa):
    x = (x - mu) / b
    kap_inv = 1 / kappa
    if x >= 0:
        ald_x = x * -kappa
    else:
        ald_x = x * kap_inv
    ald_x -= np.log(kappa + kap_inv)
    return ald_x - np.log(b)


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, b, kappa):
    return (-nb_logpdf(x, mu, b, kappa)).sum()


@nb.njit(cache=True)
def nb_rvs(random_samples, mu, b, kappa):
    sgn = np.sign(random_samples)
    return mu - (1 / (1 / b * sgn * kappa**sgn)) * np.log(1 - random_samples * sgn * kappa**sgn)


@nb.njit(cache=True)
def nb_entropy(b, kappa):
    return 1 + np.log(kappa + 1 / kappa) + np.log(b)



import numba as nb
import numpy as np

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml, optimize_moments
from preliz.internal.special import cdf_bounds, ppf_bounds_cont


class LogLogistic(Continuous):
    r"""
    Log-Logistic distribution.

    Also known as the Fisk distribution is a continuous non-negative distribution used in survival
    analysis as a parametric model for events whose rate increases initially and decreases later

    The pdf of this distribution is

    .. math::

        f(x\mid \alpha, \beta) =
            \frac{ (\beta/\alpha)(x/\alpha)^{\beta-1}}
            {\left( 1+(x/\alpha)^{\beta} \right)^2}

    .. plot::
        :context: close-figs


        from preliz import LogLogistic, style
        style.use('preliz-doc')
        alphas = [1, 1, 2]
        betas = [4, 8, 8]
        for alpha, beta in zip(alphas, betas):
            LogLogistic(alpha,beta).plot_pdf(support=(0, 6))

    ========  ==========================================================================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`{\alpha\,\pi/\beta \over \sin(\pi/\beta)}`
    Variance  :math:`\alpha^2 \left(2b / \sin 2b -b^2 / \sin^2 b \right), \quad \beta>2`
    ========  ==========================================================================

    Parameters
    ----------
    alpha : float
        Scale parameter. (alpha > 0))
    beta : float
        Shape parameter. (beta > 0)).
    """

    def __init__(self, alpha=None, beta=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, beta)

    def _parametrization(self, alpha=None, beta=None):
        self.alpha = alpha
        self.beta = beta
        self.params = (self.alpha, self.beta)
        self.param_names = ("alpha", "beta")
        self.params_support = ((eps, np.inf), (eps, np.inf))
        if all_not_none(alpha, beta):
            self._update(alpha, beta)

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.params = (self.alpha, self.beta)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta)

    def mean(self):
        return np.where(
            self.beta > 1, self.alpha * np.pi / self.beta / np.sin(np.pi / self.beta), np.nan
        )

    def mode(self):
        return np.where(
            self.beta > 1, self.alpha * ((self.beta - 1) / (self.beta + 1)) ** (1 / self.beta), 0
        )

    def median(self):
        return self.alpha

    def var(self):
        pib = np.pi / self.beta
        return np.where(
            self.beta > 2,
            self.alpha**2 * (2 * pib / np.sin(2 * pib) - pib**2 / np.sin(pib) ** 2),
            np.nan,
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        mean = self.mean()
        std = self.std()
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.where(
            self.beta > 3, np.trapz(((x_values - mean) / std) ** 3 * pdf, x_values), np.nan
        )

    def kurtosis(self):
        mean = self.mean()
        std = self.std()
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.where(
            self.beta > 4, np.trapz(((x_values - mean) / std) ** 4 * pdf, x_values) - 3, np.nan
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        u_val = random_state.random(size)
        return self.alpha * (u_val / (1 - u_val)) ** (1 / self.beta)

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, alpha, beta):
    return cdf_bounds(1 / (1 + (x / alpha) ** (-beta)), x, 0, np.inf)


@nb.njit(cache=True)
def nb_ppf(q, alpha, beta):
    return ppf_bounds_cont(alpha * (q / (1 - q)) ** (1 / beta), q, 0, np.inf)


@nb.njit(cache=True)
def nb_entropy(alpha, beta):
    return np.log(alpha) - np.log(beta) + 2


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta):
    if x <= 0:
        return -np.inf
    elif x == np.inf:
        return -np.inf
    else:
        return (
            -2 * np.log1p((x / alpha) ** beta)
            + (-1 + beta) * (np.log(x) - np.log(alpha))
            - np.log(alpha)
            + np.log(beta)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()



import numba as nb
import numpy as np
from scipy.special import gammainc, gammaincinv

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, any_not_none, eps
from preliz.internal.optimization import optimize_ml
from preliz.internal.special import cdf_bounds, digamma, gammaln, ppf_bounds_cont, xlogy


class Gamma(Continuous):
    r"""
    Gamma distribution.

    Represents the sum of alpha exponentially distributed random variables,
    each of which has rate beta.

    The pdf of this distribution is

    .. math::

        f(x \mid \alpha, \beta) =
            \frac{\beta^{\alpha}x^{\alpha-1}e^{-\beta x}}{\Gamma(\alpha)}

    .. plot::
        :context: close-figs

        from preliz import Gamma, style
        style.use('preliz-doc')
        alphas = [1., 3., 7.5]
        betas = [.5, 1., 1.]
        for alpha, beta in zip(alphas, betas):
            Gamma(alpha, beta).plot_pdf()

    ========  ===============================
    Support   :math:`x \in (0, \infty)`
    Mean      :math:`\dfrac{\alpha}{\beta}`
    Variance  :math:`\dfrac{\alpha}{\beta^2}`
    ========  ===============================

    Gamma distribution has 2 alternative parameterizations. In terms of alpha and
    beta or mu (mean) and sigma (standard deviation).

    The link between the 2 alternatives is given by

    .. math::

        \alpha &= \frac{\mu^2}{\sigma^2} \\
        \beta  &= \frac{\mu}{\sigma^2}

    Parameters
    ----------
    alpha : float
        Shape parameter (alpha > 0).
    beta : float
        Rate parameter (beta > 0).
    mu : float
        Mean (mu > 0).
    sigma : float
        Standard deviation (sigma > 0)

    """

    def __init__(self, alpha=None, beta=None, mu=None, sigma=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, beta, mu, sigma)

    def _parametrization(self, alpha=None, beta=None, mu=None, sigma=None):
        if any_not_none(alpha, beta) and any_not_none(mu, sigma):
            raise ValueError(
                "Incompatible parametrization. Either use alpha and beta or mu and sigma."
            )

        self.param_names = ("alpha", "beta")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if any_not_none(mu, sigma):
            self.mu = mu
            self.sigma = sigma
            self.param_names = ("mu", "sigma")
            if all_not_none(mu, sigma):
                alpha, beta = self._from_mu_sigma(mu, sigma)

        self.alpha = alpha
        self.beta = beta
        if all_not_none(self.alpha, self.beta):
            self._update(self.alpha, self.beta)

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.mu, self.sigma = self._to_mu_sigma(self.alpha, self.beta)

        if self.param_names[0] == "alpha":
            self.params = (self.alpha, self.beta)
        elif self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma)

        self.is_frozen = True

    def _from_mu_sigma(self, mu, sigma):
        alpha = mu**2 / sigma**2
        beta = mu / sigma**2
        return alpha, beta

    def _to_mu_sigma(self, alpha, beta):
        mu = alpha / beta
        sigma = alpha**0.5 / beta
        return mu, sigma

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.alpha, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta, 0, np.inf)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, 0, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta)

    def mean(self):
        return self.alpha / self.beta

    def mode(self):
        return np.where(self.alpha <= 1, 0, (self.alpha - 1) / self.beta)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.alpha / self.beta**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 2 / self.alpha**0.5

    def kurtosis(self):
        return 6 / self.alpha

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.gamma(self.alpha, 1 / self.beta, size)

    def _fit_moments(self, mean, sigma):
        alpha, beta = self._from_mu_sigma(mean, sigma)
        self._update(alpha, beta)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


def nb_cdf(x, alpha, beta, lower, upper):
    prob = gammainc(alpha, x / (1 / beta))
    return cdf_bounds(prob, x, lower, upper)


def nb_ppf(q, alpha, beta, lower, upper):
    x_val = gammaincinv(alpha, q) * (1 / beta)
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta):
    if x <= 0:
        return -np.inf
    else:
        x = x / (1 / beta)
        return xlogy(alpha - 1.0, x) - x - gammaln(alpha) - np.log(1 / beta)


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()


@nb.njit(cache=True)
def nb_entropy(alpha, beta):
    return alpha - np.log(beta) + gammaln(alpha) + (1 - alpha) * digamma(alpha)



"""Optimization routines and utilities."""

import warnings
from copy import copy

import numpy as np
from scipy.optimize import brentq, least_squares, minimize, minimize_scalar, root_scalar
from scipy.special import i0, i0e, i1, i1e, logit

from preliz.internal.distribution_helper import init_vals as default_vals


def optimize_max_ent(dist, lower, upper, mass, none_idx, fixed_params, fixed_stat):
    def prob_bound(params, dist, lower, upper, mass):
        params = get_params(dist, params, none_idx, fixed_params)
        dist._parametrization(**params)
        if dist.kind == "discrete":
            lower -= 1
        cdf0 = dist.cdf(lower)
        cdf1 = dist.cdf(upper)
        loss = (cdf1 - cdf0) - mass
        if fixed_stat:
            loss = loss - abs(getattr(dist, fixed_stat[0])() - fixed_stat[1])
        return loss

    def entropy_loss(params, dist):
        params = get_params(dist, params, none_idx, fixed_params)
        dist._parametrization(**params)
        return -dist.entropy()

    cons = {
        "type": "eq",
        "fun": prob_bound,
        "args": (dist, lower, upper, mass),
    }
    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Values in x were outside bounds")
        opt = minimize(entropy_loss, x0=init_vals, bounds=bounds, args=(dist), constraints=cons)

    params = get_params(dist, opt["x"], none_idx, fixed_params)
    dist._parametrization(**params)

    return opt


def get_params(dist, params, none_idx, fixed):
    params_ = {}
    pdx = 0
    fdx = 0
    for idx in range(len(dist.params)):
        name = dist.param_names[idx]
        if idx in none_idx:
            params_[name] = params[pdx]
            pdx += 1
        else:
            params_[name] = fixed[fdx]
            fdx += 1

    return params_


def optimize_quartile(dist, x_vals, none_idx, fixed):
    def func(params, dist, x_vals):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = dist.cdf(x_vals) - [0.25, 0.5, 0.75]
        return loss

    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    bounds = list(zip(*bounds))

    opt = least_squares(func, x0=init_vals, args=(dist, x_vals), bounds=bounds)
    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    return opt


def optimize_pdf(dist, x_vals, epdf, none_idx, fixed):
    def func(params, dist, x_vals, epdf):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = dist.pdf(x_vals) - epdf
        return loss

    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    bounds = list(zip(*bounds))

    opt = least_squares(func, x0=init_vals, args=(dist, x_vals, epdf), bounds=bounds)
    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    loss = opt["cost"]
    return loss


def optimize_moments(dist, mean, sigma, params=None):
    def func(params, dist, mean, sigma):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = abs(dist.mean() - mean) + abs(dist.std() - sigma)
        return loss

    none_idx, fixed = get_fixed_params(dist)

    if params is not None:
        dist._update(*params)
    else:
        name = dist.__class__.__name__
        if name == "Truncated":
            vals = copy(default_vals["Truncated"])
            vals.update(default_vals[dist.dist.__class__.__name__])
            dist._parametrization(**vals)
        else:
            dist._update(**default_vals[name])

    init_vals = np.array(dist.params)[none_idx]

    if dist.__class__.__name__ in ["HyperGeometric", "BetaBinomial"]:
        opt = least_squares(func, x0=init_vals, args=(dist, mean, sigma))
    else:
        bounds = np.array(dist.params_support)[none_idx]
        bounds = list(zip(*bounds))
        if dist.__class__.__name__ in ["DiscreteWeibull"]:
            opt = least_squares(
                func, x0=init_vals, args=(dist, mean, sigma), bounds=bounds, loss="soft_l1"
            )
        else:
            opt = least_squares(func, x0=init_vals, args=(dist, mean, sigma), bounds=bounds)

    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    return opt


def optimize_moments_rice(mean, std_dev):
    """
    Moment matching for the Rice distribution.

    This function uses the Koay inversion technique, see: https://doi.org/10.1016/j.jmr.2006.01.016
    and https://en.wikipedia.org/wiki/Rice_distribution
    """
    ratio = mean / std_dev

    if ratio < 1.913:  # Rayleigh distribution
        nu = np.finfo(float).eps
        sigma = 0.655 * std_dev
    else:

        def xi(theta):
            return (
                2
                + theta**2
                - np.pi
                / 8
                * np.exp(-(theta**2) / 2)
                * ((2 + theta**2) * i0(theta**2 / 4) + theta**2 * i1(theta**2 / 4)) ** 2
            )

        def fpf(theta):
            return (xi(theta) * (1 + ratio**2) - 2) ** 0.5

        def func(theta):
            return np.abs(fpf(theta) - theta)

        theta = minimize(func, x0=fpf(1), bounds=[(0, None)]).x
        xi_theta = xi(theta)
        sigma = std_dev / xi_theta**0.5
        nu = (mean**2 + (xi_theta - 2) * sigma**2) ** 0.5

    return nu, sigma


def optimize_ml(dist, sample):
    def negll(params, dist, sample):
        dist._update(*params)
        return dist._neg_logpdf(sample)

    dist._fit_moments(np.mean(sample), np.std(sample))
    init_vals = dist.params

    opt = minimize(negll, x0=init_vals, bounds=dist.params_support, args=(dist, sample))

    dist._update(*opt["x"])

    return opt


def optimize_dirichlet_mode(lower_bounds, mode, target_mass, _dist):
    def prob_approx(tau, lower_bounds, mode, _dist):
        alpha = [1 + tau * mode_i for mode_i in mode]
        a_0 = sum(alpha)
        marginal_prob_list = []
        for a_i, lbi in zip(alpha, lower_bounds):
            _dist._parametrization(a_i, a_0 - a_i)
            marginal_prob_list.append(_dist.cdf(lbi))

        mean_cdf = np.mean(marginal_prob_list)
        return mean_cdf, alpha

    tau = 1
    new_prob, alpha = prob_approx(tau, lower_bounds, mode, _dist)

    while abs(new_prob - target_mass) > 0.0001:
        if new_prob < target_mass:
            tau -= 0.5 * tau
        else:
            tau += 0.5 * tau

        new_prob, alpha = prob_approx(tau, lower_bounds, mode, _dist)

    return new_prob, alpha


def optimize_beta_mode(lower, upper, tau_not, mode, dist, mass, prob):
    while abs(prob - mass) > 0.0001:
        alpha = 1 + mode * tau_not
        beta = 1 + (1 - mode) * tau_not
        dist._parametrization(alpha, beta)
        prob = dist.cdf(upper) - dist.cdf(lower)

        if prob > mass:
            tau_not -= 0.5 * tau_not
        else:
            tau_not += 0.5 * tau_not


def optimize_hdi(dist, mass):
    def interval_loss(params):
        cdf = dist.cdf(params)
        loss = (cdf[1] - cdf[0] + mass_at_boundaries) - mass
        return loss

    def interval_short(params):
        if params[1] < params[0]:
            return np.inf
        return params[1] - params[0]

    cons = {
        "type": "eq",
        "fun": interval_loss,
    }
    lower, upper = dist.support
    mass_at_boundaries = dist.pdf(lower) + dist.pdf(upper)
    init_vals = dist.eti(mass=mass, fmt="none")
    bounds = [(lower, upper)]
    opt = minimize(interval_short, x0=init_vals, bounds=bounds, constraints=cons)

    lower, upper = opt.x
    if dist.kind == "discrete":
        upper = np.floor(upper - 1).astype(int)
        lower = np.ceil(lower).astype(int)

    return lower, upper


def optimize_pymc_model(
    fmodel,
    target,
    num_draws,
    opt_iterations,
    initial_guess,
    rng,
):
    prior_array = np.zeros((opt_iterations, len(initial_guess)))

    for idx in range(opt_iterations + 1):
        # can we sample systematically from these and less random?
        # This should be more flexible and allow other targets than just
        # a PreliZ distribution
        if isinstance(target, list):
            obs = get_weighted_rvs(target, num_draws, rng)
        else:
            obs = target.rvs(num_draws, idx)
        result = minimize(
            fmodel,
            initial_guess,
            tol=0.001,
            method="powell",
            args=(obs),
        )
        # To help minimize the effect of priors we don't save the first result
        # and instead we use it as the initial guess for the next optimization
        # Updating the initial guess also helps reduce computational times
        if idx:
            prior_array[idx - 1] = result.x
        initial_guess = result.x

    return prior_array


def relative_error(dist, lower, upper, required_mass):
    if dist.kind == "discrete":
        lower -= 1
    computed_mass = dist.cdf(upper) - dist.cdf(lower)
    return abs((computed_mass - required_mass) / required_mass * 100), computed_mass


def fit_to_epdf(selected_distributions, x_vals, epdf, mean, std, x_min, x_max, extra_pros):
    """
    Minimize the difference between the pdf and the epdf.

    Minimization is done over a grid of values defined by x_min and x_max. This function is
    intended to be used with pz.roulette
    """
    fitted = Loss(len(selected_distributions))
    for dist in selected_distributions:
        if dist.__class__.__name__ in extra_pros:
            try:
                dist._parametrization(**extra_pros[dist.__class__.__name__])
            except TypeError:
                pass
        if dist.__class__.__name__ == "BetaScaled":
            update_bounds_beta_scaled(dist, x_min, x_max)

        if dist._check_endpoints(x_min, x_max, raise_error=False):
            none_idx, fixed = get_fixed_params(dist)
            dist._fit_moments(mean, std)
            loss = optimize_pdf(dist, x_vals, epdf, none_idx, fixed)

            fitted.update(loss, dist)

    return fitted.dist


def fit_to_sample(selected_distributions, sample, x_min, x_max):
    """Maximize the likelihood given a sample."""
    fitted = Loss(len(selected_distributions))
    for dist in selected_distributions:
        if dist.__class__.__name__ in ["BetaScaled", "TruncatedNormal"]:
            update_bounds_beta_scaled(dist, x_min, x_max)

        loss = np.inf
        if dist._check_endpoints(x_min, x_max, raise_error=False):
            if sample.ndim > 1:
                dists = []
                neg_logpdf = 0
                for s in sample:
                    dist_i = copy(dist)
                    dist_i._fit_mle(s)
                    neg_logpdf += dist_i._neg_logpdf(s)
                    dists.append(dist_i)
                new_dict = {}
                for d in dists:
                    params = d.params_dict
                    for k, v in params.items():
                        if k in new_dict:
                            new_dict[k].append(v)
                        else:
                            new_dict[k] = [v]

                dist._parametrization(**{k: np.asarray(v) for k, v in new_dict.items()})
            else:
                dist._fit_mle(sample)
                neg_logpdf = dist._neg_logpdf(sample)
            corr = get_penalization(sample.size, dist)
            loss = neg_logpdf + corr
        fitted.update(loss, dist)

    return fitted


def fit_to_quartile(selected_distributions, q1, q2, q3, extra_pros):
    error = np.inf
    fitted_dist = None

    for distribution in selected_distributions:
        if distribution.__class__.__name__ in extra_pros:
            distribution._parametrization(**extra_pros[distribution.__class__.__name__])
            if distribution.__class__.__name__ == "BetaScaled":
                update_bounds_beta_scaled(
                    distribution,
                    extra_pros[distribution.__class__.__name__]["lower"],
                    extra_pros[distribution.__class__.__name__]["upper"],
                )
        if distribution._check_endpoints(q1, q3, raise_error=False):
            none_idx, fixed = get_fixed_params(distribution)

            distribution._fit_moments(mean=q2, sigma=(q3 - q1) / 1.35)

            optimize_quartile(distribution, (q1, q2, q3), none_idx, fixed)

            r_error, _ = relative_error(distribution, q1, q3, 0.5)
            if r_error < error:
                fitted_dist = distribution
                error = r_error

    return fitted_dist


def update_bounds_beta_scaled(dist, x_min, x_max):
    dist.lower = x_min
    dist.upper = x_max
    dist.support = (x_min, x_max)
    return dist


def get_penalization(n, dist):
    """
    AIC with a correction for small sample sizes.

    Burnham, K. P.; Anderson, D. R. (2004),
    "Multimodel inference: understanding AIC and BIC in Model Selection"
    https://doi.org/10.1177/0049124104268
    """
    k = len(dist.params)
    return k + ((k + 1) * k) / (-k + n - 1)


class Loss:
    def __init__(self, size=None):
        self.old_loss = np.inf
        self.dist = None
        self.count = 0
        self.distributions = np.empty(size, dtype=object)
        self.losses = np.empty(size)

    def update(self, loss, dist):
        self.distributions[self.count] = dist
        self.losses[self.count] = loss
        self.count += 1
        if loss < self.old_loss:
            self.old_loss = loss
            self.dist = dist


def get_fixed_params(distribution):
    none_idx = []
    fixed = []
    for idx, p_n in enumerate(distribution.param_names):
        value = getattr(distribution, p_n)
        if value is None:
            none_idx.append(idx)
        else:
            fixed.append(value)
    return none_idx, fixed


def find_kappa(data, mu):
    ere = np.mean(np.cos(mu - data))

    if ere > 0:

        def solve_for_kappa(kappa):
            return i1e(kappa) / i0e(kappa) - ere

        root_res = root_scalar(
            solve_for_kappa, method="brentq", bracket=(np.finfo(float).tiny, 1e16)
        )
        return root_res.root
    else:
        return np.finfo(float).tiny


def find_mode_logitnormal(distribution):
    eps = np.finfo(float).eps

    def mode_equation(x):
        # The equation is: logit(x) = σ²(2x-1) + μ
        # We want to find the root of: logit(x) - σ²(2x-1) - μ = 0
        return logit(x) - (distribution.sigma**2 * (2 * x - 1)) - distribution.mu

    # Left side
    try:
        sol1 = root_scalar(mode_equation, bracket=(eps, 0.5 - eps)).root
    except ValueError:
        sol1 = None

    # Right side
    try:
        sol2 = root_scalar(mode_equation, bracket=(0.5 + eps, 1 - eps)).root
    except ValueError:
        sol2 = None

    if sol1 is None and sol2 is None:
        # If no solutions found, return the median as an approximation
        return distribution.median()
    elif sol1 is None:
        return sol2
    elif sol2 is None:
        return sol1
    else:
        # Return the solution with higher density
        if distribution.pdf(sol1) >= distribution.pdf(sol2):
            return sol1
        return sol2


def find_mode(distribution):
    """
    Find mode of a distribution through numerical optimization.

    Parameters
    ----------
    distribution : Distribution
        Distribution object that has pdf method

    Returns
    -------
    float
        Mode of the distribution
    """

    def negative_pdf(x):
        return -distribution.pdf(x)

    bounds = distribution._finite_endpoints("full")
    result = minimize_scalar(negative_pdf, bounds=bounds, method="bounded")
    return result.x


def find_discrete_mode(dist):
    """
    Find mode for a discrete distribution from its pmf.

    Parameters
    ----------
    dist : Distribution
        PreliZ distribution object

    Returns
    -------
    int
        Mode of the distribution
    """
    x_vals = dist.xvals("full")
    pmf_vals = dist.pdf(x_vals)
    return int(x_vals[np.argmax(pmf_vals)])


def find_ppf(dist, q):
    q = np.atleast_1d(q)
    ppf = np.zeros_like(q)
    lower, upper = dist.support
    for idx, q_i in enumerate(q):
        if q_i < 0:
            ppf[idx] = np.nan
        elif q_i > 1:
            ppf[idx] = np.nan
        elif q_i == 0:
            if dist.kind == "discrete":
                ppf[idx] = lower - 1
            else:
                ppf[idx] = lower
        elif q_i == 1:
            ppf[idx] = upper
        elif dist.__class__.__name__ in ["HyperGeometric", "BetaBinomial"]:
            ppf[idx] = _ppf_single(dist, q_i) + 1
        else:
            ppf[idx] = _ppf_single(dist, q_i)
    return ppf[0] if len(ppf) == 1 else ppf


def _ppf_single(dist, q):
    def func(x, dist, q):
        return dist.cdf(x) - q

    factor = 10.0
    left, right = dist.support

    left = min(-factor, right)
    while func(left, dist, q) > 0.0:
        left, right = left * factor, left

    right = max(factor, left)
    while func(right, dist, q) < 0.0:
        left, right = right, right * factor

    return brentq(func, left, right, args=(dist, q))


def get_weighted_rvs(target, size, rng):
    targets = [t[0] for t in target]
    weights = [t[1] for t in target]
    target_rnd_choices = np.random.choice(len(targets), size=size, p=weights)
    samples = [target.rvs(size, random_state=rng) for target in targets]
    return np.choose(target_rnd_choices, samples)


def _root(n_p, k_sq, a_sq, x):
    def _fixed_point(t_s, x_len, k_sq, a_sq):
        """Calculate t-zeta*gamma^[l](t).

        Implementation of the function t-zeta*gamma^[l](t) derived from equation (30) in [1].

        References
        ----------
        .. [1] Kernel density estimation via diffusion.
        Z. I. Botev, J. F. Grotowski, and D. P. Kroese.
        Ann. Statist. 38 (2010), no. 5, 2916--2957.
        """
        k_sq = np.asarray(k_sq, dtype=np.float64)
        a_sq = np.asarray(a_sq, dtype=np.float64)

        k_o = 7
        func = np.sum(np.power(k_sq, k_o) * a_sq * np.exp(-k_sq * np.pi**2 * t_s))
        func *= 0.5 * np.pi ** (2.0 * k_o)

        for ite in np.arange(k_o - 1, 2 - 1, -1):
            c_1 = (1 + 0.5 ** (ite + 0.5)) / 3
            c_2 = np.prod(np.arange(1.0, 2 * ite + 1, 2, dtype=np.float64))
            c_2 /= (np.pi / 2) ** 0.5
            t_j = np.power((c_1 * (c_2 / (x_len * func))), (2.0 / (3.0 + 2.0 * ite)))
            func = np.sum(k_sq**ite * a_sq * np.exp(-k_sq * np.pi**2.0 * t_j))
            func *= 0.5 * np.pi ** (2 * ite)

        out = t_s - (2 * x_len * np.pi**0.5 * func) ** (-0.4)
        return out

    # The right bound is at most 0.01
    found = False
    n_p = max(min(1050, n_p), 50)
    tol = 10e-12 + 0.01 * (n_p - 50) / 1000

    while not found:
        try:
            band_w, res = brentq(
                _fixed_point, 0, 0.01, args=(n_p, k_sq, a_sq), full_output=True, disp=False
            )
            found = res.converged
        except ValueError:
            band_w = 0
            tol *= 2.0
            found = False
        if band_w <= 0 or tol >= 1:
            band_w = (_bw_silverman(x) / np.ptp(x)) ** 2
            return band_w
    return band_w


def _bw_silverman(x, x_std=None):
    """Silverman's Rule."""
    x_std = np.std(x)
    q75, q25 = np.percentile(x, [75, 25])
    x_iqr = q75 - q25
    a = min(x_std, x_iqr / 1.34)
    band_w = 0.9 * a * len(x) ** (-0.2)
    return band_w



import matplotlib.pyplot as plt
import pytest
from numpy.testing import assert_allclose

import preliz as pz
from preliz.distributions import (
    AsymmetricLaplace,
    Bernoulli,
    Beta,
    BetaBinomial,
    BetaScaled,
    Binomial,
    Cauchy,
    ChiSquared,
    DiscreteUniform,
    DiscreteWeibull,
    ExGaussian,
    Exponential,
    Gamma,
    Geometric,
    Gumbel,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    HyperGeometric,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogitNormal,
    LogLogistic,
    LogNormal,
    Moyal,
    NegativeBinomial,
    Normal,
    Pareto,
    Poisson,
    Rice,
    SkewNormal,
    SkewStudentT,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
    ZeroInflatedBinomial,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
)


@pytest.mark.parametrize(
    "distribution, params",
    [
        (AsymmetricLaplace, (2, 3, 1)),
        (Beta, (2, 5)),
        (BetaScaled, (2, 5, -1, 4)),
        (Cauchy, (0, 1)),
        (ChiSquared, (5,)),
        (ExGaussian, (0, 1, 3)),
        (Exponential, (5,)),
        (Gamma, (2, 5)),
        (Gumbel, (0, 2)),
        (HalfCauchy, (1,)),
        (HalfNormal, (1,)),
        (HalfStudentT, (5, 1)),
        (HalfNormal, (2,)),
        (InverseGamma, (3, 5)),
        (Kumaraswamy, (2, 3)),
        (Laplace, (0, 2)),
        (Logistic, (0, 1)),
        (LogLogistic, (1, 5)),
        (LogNormal, (0, 1)),
        (LogitNormal, (0, 0.5)),
        (Moyal, (0, 2)),
        (Normal, (0, 1)),
        (Pareto, (5, 1)),
        (Rice, (0, 2)),
        (SkewNormal, (0, 1, -6)),
        (SkewStudentT, (0, 1, 2, 2)),
        (StudentT, (4, 0, 1)),
        (Triangular, (0, 3, 4)),
        (TruncatedNormal, (0, 0.5, -1, 1)),
        (Uniform, (2, 5)),
        (VonMises, (1, 2)),
        (Wald, (2, 1)),
        (Weibull, (2, 1)),
        (Bernoulli, (0.5,)),
        (BetaBinomial, (2, 5, 10)),
        (Binomial, (5, 0.5)),
        (DiscreteUniform, (-2, 2)),
        (DiscreteWeibull, (0.9, 1.3)),
        (Geometric, (0.75,)),
        (HyperGeometric, (50, 10, 20)),
        (NegativeBinomial, (10, 2.5)),
        (Poisson, (4.2,)),
        (ZeroInflatedBinomial, (0.5, 10, 0.6)),
        (ZeroInflatedNegativeBinomial, (0.7, 8, 4)),
        (
            ZeroInflatedPoisson,
            (
                0.8,
                4.2,
            ),
        ),
    ],
)
def test_auto_recover(distribution, params):
    for _ in range(10):
        sample = distribution(*params).rvs(20_000)
        dist = distribution()
        try:
            if dist.__class__.__name__ in [
                "BetaScaled",
                "TruncatedNormal",
            ]:
                tol = 1
            else:
                tol = 0.5
            pz.mle([dist], sample, plot=0)
            assert_allclose(dist.params, params, atol=tol)
            break
        except AssertionError:
            pass
    else:
        raise AssertionError(f"Test failed after 10 attempts.{dist.params}")


def test_recover_right():
    dists = [Normal(), Gamma(), Poisson()]
    sample = Normal(0, 1).rvs(10000)
    idx, ax = pz.mle(dists, sample, plot=0)
    assert len(idx) == len(dists)
    assert idx[0] == 0
    assert ax is None

    dists = [pz.Normal(), pz.Gamma(), pz.Poisson()]
    sample = pz.Normal(10, 0.5).rvs((2, 10000), random_state=123)
    idx, ax = pz.mle(dists, sample, plot=0)
    all(d.params[0].size == 2 for d in dists)
    assert idx[0] == 0
    assert ax is None

    plt.figure()
    sample = Gamma(2, 10).rvs(10000)
    idx, ax = pz.mle(dists, sample)
    assert idx[0] == 1
    assert len(ax.get_legend().legend_handles) == 1

    plt.figure()
    sample = Poisson(10).rvs(10000)
    idx, ax = pz.mle(dists, sample, plot=3)
    assert idx[0] == 2
    assert len(ax.get_legend().legend_handles) == 3



# Continuous Distributions
from preliz.distributions.asymmetric_laplace import AsymmetricLaplace

# Discrete Distributions
from preliz.distributions.bernoulli import Bernoulli
from preliz.distributions.beta import Beta
from preliz.distributions.betabinomial import BetaBinomial
from preliz.distributions.betascaled import BetaScaled
from preliz.distributions.binomial import Binomial
from preliz.distributions.categorical import Categorical
from preliz.distributions.cauchy import Cauchy
from preliz.distributions.censored import Censored
from preliz.distributions.chi_squared import ChiSquared

# Multivariate Distributions
from preliz.distributions.continuous_multivariate import Dirichlet, MvNormal
from preliz.distributions.discrete_uniform import DiscreteUniform
from preliz.distributions.discrete_weibull import DiscreteWeibull
from preliz.distributions.exgaussian import ExGaussian
from preliz.distributions.exponential import Exponential
from preliz.distributions.gamma import Gamma
from preliz.distributions.geometric import Geometric
from preliz.distributions.gumbel import Gumbel
from preliz.distributions.halfcauchy import HalfCauchy
from preliz.distributions.halfnormal import HalfNormal
from preliz.distributions.halfstudentt import HalfStudentT
from preliz.distributions.hurdle import Hurdle
from preliz.distributions.hypergeometric import HyperGeometric
from preliz.distributions.inversegamma import InverseGamma
from preliz.distributions.kumaraswamy import Kumaraswamy
from preliz.distributions.laplace import Laplace
from preliz.distributions.logistic import Logistic
from preliz.distributions.logitnormal import LogitNormal
from preliz.distributions.loglogistic import LogLogistic
from preliz.distributions.lognormal import LogNormal

# Transform Distributions
from preliz.distributions.mixture import Mixture
from preliz.distributions.moyal import Moyal
from preliz.distributions.negativebinomial import NegativeBinomial
from preliz.distributions.normal import Normal
from preliz.distributions.pareto import Pareto
from preliz.distributions.poisson import Poisson
from preliz.distributions.rice import Rice
from preliz.distributions.skew_studentt import SkewStudentT
from preliz.distributions.skewnormal import SkewNormal
from preliz.distributions.studentt import StudentT
from preliz.distributions.triangular import Triangular
from preliz.distributions.truncated import Truncated
from preliz.distributions.truncatednormal import TruncatedNormal
from preliz.distributions.uniform import Uniform
from preliz.distributions.vonmises import VonMises
from preliz.distributions.wald import Wald
from preliz.distributions.weibull import Weibull
from preliz.distributions.zi_binomial import ZeroInflatedBinomial
from preliz.distributions.zi_negativebinomial import ZeroInflatedNegativeBinomial
from preliz.distributions.zi_poisson import ZeroInflatedPoisson

all_continuous = [
    AsymmetricLaplace,
    Beta,
    BetaScaled,
    Cauchy,
    ChiSquared,
    ExGaussian,
    Exponential,
    Gamma,
    Gumbel,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogLogistic,
    LogNormal,
    LogitNormal,
    Moyal,
    Normal,
    Pareto,
    Rice,
    SkewNormal,
    SkewStudentT,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
]
all_discrete = [
    Bernoulli,
    BetaBinomial,
    Binomial,
    Categorical,
    DiscreteUniform,
    DiscreteWeibull,
    Geometric,
    HyperGeometric,
    NegativeBinomial,
    Poisson,
    ZeroInflatedBinomial,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
]

all_continuous_multivariate = [Dirichlet, MvNormal]


__all__ = (  # noqa: PLE0604
    [s.__name__ for s in all_continuous]
    + [s.__name__ for s in all_discrete]
    + [s.__name__ for s in all_continuous_multivariate]
    + [Mixture.__name__]
    + [Truncated.__name__]
    + [Censored.__name__]
    + [Hurdle.__name__]
)



from preliz.distributions import Gamma, Normal
from preliz.unidimensional.combine import combine


def test_combine():
    # Create some sample distributions
    distributions = [Normal(0, 1), Gamma(2, 1)]

    # Call the combine function
    fit_dists, ax = combine(distributions, rng=413)

    assert [dist.__class__.__name__ for dist in fit_dists[:2]] == ["StudentT", "Normal"]
    assert [dist.is_frozen for dist in fit_dists] == [True, True, False, False]

    fit_dists, ax = combine(distributions, weights=[0, 1], rng=None, plot=0)

    assert [dist.__class__.__name__ for dist in fit_dists[:2]] == ["Gamma", "LogNormal"]
    assert [dist.is_frozen for dist in fit_dists] == [True, True, True, True]
    assert ax is None

    fit_dists, _ = combine(distributions, dist_names=["Moyal"])
    assert len(fit_dists) == 1
    assert fit_dists[0].__class__.__name__ == "Moyal"



import numpy as np

from preliz.internal.distribution_helper import get_distributions
from preliz.unidimensional.mle import mle


def combine(
    distributions,
    weights=None,
    dist_names=None,
    sample_size=10_000,
    rng=0,
    plot=1,
    plot_kwargs=None,
    ax=None,
):
    """
    Combine a set of distributions into a single one.

    Fit a weighted sample from ``distributions`` into the distributions listed in ``dist_names`.
    The fit is done using maximum likelihood estimation, and the best match is plotted.
    Notice that the result is NOT a Mixture distribution, but a single distribution
    that best fits the weighted sample.

    Parameters
    ----------
    distributions : List of PreliZ distributions
        These are the distributions that we want to combine. Typically, these have been
        elicited from different individuals or instances.
    weights : array-like, optional
        Weights for each distribution. Defaults to None, i.e. equal weights.
        The sum of the weights must be equal to 1, otherwise it will be normalized.
    dist_names: list
        List of distributions to fit the weighted sample.
        Defaults to ``["Normal", "Gamma", "LogNormal", "StudentT"]``.
    sample_size : int
        Number of total samples to generate for the fit.
    rng : int or numpy.random.Generator, optional
        Random number generator or seed. Defaults to ``0``.
    plot : int
        Number of distributions to plots. Defaults to ``1`` (i.e. plot the best match)
        If larger than the number of passed distributions it will plot all of them.
        Use ``0`` or ``False`` to disable plotting.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()``.
    ax : matplotlib axes

    Returns
    -------
    fitted_distributions : list of PreliZ distributions.
        The distributions that best fit the weighted sample. Sorted from best to worst match.
    axes : matplotlib axes
    """
    if rng is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(rng)

    if weights is None:
        weights = np.full(len(distributions), 1 / len(distributions))
    else:
        weights = np.array(weights, dtype=float)

    if np.any(weights < 0):
        raise ValueError("The weights must be positive.")

    weights /= weights.sum()

    n_size = (sample_size * weights).astype(int)

    if dist_names is None:
        dist_names = ["Normal", "Gamma", "LogNormal", "StudentT"]

    sample = []
    for dist, n in zip(distributions, n_size):
        sample.append(dist.rvs(n, random_state=rng))

    distributions = get_distributions(dist_names)

    idx, ax = mle(distributions, np.concatenate(sample), plot=plot, plot_kwargs=plot_kwargs, ax=ax)

    return np.array(distributions)[idx], ax



import numba as nb
import numpy as np
from scipy.special import comb

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps, from_precision, to_precision
from preliz.internal.optimization import optimize_ml, optimize_moments
from preliz.internal.special import beta, betainc, betaincinv, cdf_bounds, gamma, ppf_bounds_cont


class SkewStudentT(Continuous):
    r"""
    Jones and Faddy's Skewed Student-t distribution.

    The pdf of this distribution is

    .. math::

        f(t)=f(t ; a, b)=C_{a, b}^{-1}\left\{1+\frac{t}{\left(a+b+t^2\right)^{1 / 2}}
        \right\}^{a+1 / 2}\left\{1-\frac{t}{\left(a+b+t^2\right)^{1 / 2}}\right\}^{b+1 / 2}

    where

    .. math::

        C_{a, b}=2^{a+b-1} B(a, b)(a+b)^{1 / 2}

    .. plot::
        :context: close-figs


        from preliz import SkewStudentT, style
        style.use('preliz-doc')
        mus = [2., 2., 4.]
        sigmas = [1., 2., 2.]
        a_s = [1., 1., 2.]
        b_s = [2., 3., 3.]
        for mu, sigma, a, b in zip(mus, sigmas, a_s, b_s):
            SkewStudentT(mu, sigma, a, b).plot_pdf(support=(-10,6))

    ========  ==========================================
    Support   :math:`x \in \mathbb{R}`
    Mean      .. math::
                  \mu + \sigma\frac{(a-b) \sqrt{(a+b)}}{2}\frac{\Gamma\left(a-\frac{1}{2}\right)
                  \Gamma\left(b-\frac{1}{2}\right)}{\Gamma(a) \Gamma(b)}
    ========  ==========================================

    Parameters
    ----------
    mu : float
        Location Parameter
    sigma : float
        Scale parameter (sigma > 0). Converges to the standard deviation as a and b approach close
    a : float
        First Shape Parameter (a > 0)
    b : float
        Second Shape Parameter (b > 0)
    lam : float
        Scale Parameter (lam > 0). Converges to the precision as a and b approach close

    Notes
    -----
    If a > b, the distribution is positively skewed, and if a < b, the distribution
    is negatively skewed. If a = b, the distribution is StudentT with
    nu (degrees of freedom) = 2a.

    """

    def __init__(self, mu=None, sigma=None, a=None, b=None, lam=None):
        super().__init__()
        self._parametrization(mu, sigma, a, b, lam)

    def _parametrization(self, mu=None, sigma=None, a=None, b=None, lam=None):
        if all_not_none(sigma, lam):
            raise ValueError("Incompatible parametrization. Either use sigma or lam.")
        self.mu = mu
        self.sigma = sigma
        self.a = a
        self.b = b
        self.support = (-np.inf, np.inf)
        self.param_names = ("mu", "sigma", "a", "b")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf), (eps, np.inf), (eps, np.inf))
        if lam is not None:
            self.lam = lam
            sigma = from_precision(lam)
            self.param_names = ("mu", "lam", "a", "b")
        if all_not_none(mu, sigma, a, b):
            self._update(mu, sigma, a, b)

    def _update(self, mu, sigma, a, b):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.lam = to_precision(sigma)
        self.a = np.float64(a)
        self.b = np.float64(b)
        if self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma, self.a, self.b)
        elif self.param_names[1] == "lam":
            self.params = (self.mu, self.lam, self.a, self.b)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.sigma, self.a, self.b))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma, self.a, self.b, -np.inf, np.inf)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma, self.a, self.b, -np.inf, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma, self.a, self.b)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma, self.a, self.b)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return (
            (self.a - self.b) * (self.a + self.b) ** 0.5 * gamma(self.a - 0.5) * gamma(self.b - 0.5)
        ) / (2 * gamma(self.a) * gamma(self.b)) * self.sigma + self.mu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        nu = (
            (self.a + self.b) ** 0.5
            / (2 * beta(self.a, self.b))
            * (
                comb(1, np.arange(2))
                * np.where(np.arange(1 + 1) % 2 > 0, -1, 1)
                * beta(self.a + 0.5 - np.arange(2), self.b - 0.5 + np.arange(2))
            ).sum()
        )
        return (
            np.where(
                np.isfinite(nu),
                (self.a + self.b)
                / (4 * beta(self.a, self.b))
                * (
                    comb(2, np.arange(3))
                    * np.where(np.arange(3) % 2 > 0, -1, 1)
                    * beta(self.a + 1 - np.arange(3), self.b - 1 + np.arange(3))
                ).sum()
                - nu**2,
                np.inf,
            )
            * self.sigma**2
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        nu1 = (
            (self.a + self.b) ** 0.5
            / (2 * beta(self.a, self.b))
            * (
                comb(1, np.arange(2))
                * np.where(np.arange(2) % 2 > 0, -1, 1)
                * beta(self.a + 0.5 - np.arange(2), self.b - 0.5 + np.arange(2))
            ).sum()
        )
        nu2 = ((self.a + self.b) / (4 * beta(self.a, self.b))) * (
            comb(2, np.arange(3))
            * np.where(np.arange(3) % 2 > 0, -1, 1)
            * beta(self.a + 1 - np.arange(3), self.b - 1 + np.arange(3))
        ).sum() - nu1**2
        nu3 = (
            ((self.a + self.b) ** 1.5 / (8 * beta(self.a, self.b)))
            * (
                comb(3, np.arange(4))
                * np.where(np.arange(4) % 2 > 0, -1, 1)
                * beta(self.a + 1.5 - np.arange(4), self.b - 1.5 + np.arange(4))
            ).sum()
            - nu1**3
            - 3 * nu1 * nu2
        )
        return nu3 / nu2**1.5

    def kurtosis(self):
        nu1 = ((self.a + self.b) ** 0.5 / (2 * beta(self.a, self.b))) * (
            comb(1, np.arange(2))
            * np.where(np.arange(2) % 2 > 0, -1, 1)
            * beta(self.a + 0.5 - np.arange(2), self.b - 0.5 + np.arange(2))
        ).sum()
        nu2 = ((self.a + self.b) / (4 * beta(self.a, self.b))) * (
            comb(2, np.arange(3))
            * np.where(np.arange(3) % 2 > 0, -1, 1)
            * beta(self.a + 1 - np.arange(3), self.b - 1 + np.arange(3))
        ).sum() - nu1**2
        nu3 = (
            ((self.a + self.b) ** 1.5 / (8 * beta(self.a, self.b)))
            * (
                comb(3, np.arange(4))
                * np.where(np.arange(4) % 2 > 0, -1, 1)
                * beta(self.a + 1.5 - np.arange(4), self.b - 1.5 + np.arange(4))
            ).sum()
            - nu1**3
            - 3 * nu1 * nu2
        )
        nu4 = (
            ((self.a + self.b) ** 2 / (16 * beta(self.a, self.b)))
            * (
                comb(4, np.arange(5))
                * np.where(np.arange(5) % 2 > 0, -1, 1)
                * beta(self.a + 2 - np.arange(5), self.b - 2 + np.arange(5))
            ).sum()
            - 4 * nu3 * nu1
            - nu1**4
            - 6 * nu2 * nu1**2
        )
        return nu4 / nu2**2 - 3

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        beta_rng = random_state.beta(self.a, self.b, size)
        return ((2 * beta_rng - 1) * np.sqrt(self.a + self.b)) / (
            2 * np.sqrt(beta_rng * (1 - beta_rng))
        ) * self.sigma + self.mu

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, mu, sigma, a, b, lower, upper):
    x = (x - mu) / sigma
    if x == -np.inf:
        return 0
    return cdf_bounds(betainc(a, b, (1 + x / np.sqrt(a + b + x**2)) * 0.5), x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma, a, b, lower, upper):
    x_val = betaincinv(a, b, q)
    return ppf_bounds_cont(
        ((2 * x_val - 1) * np.sqrt(a + b)) / (2 * np.sqrt(x_val * (1 - x_val))) * sigma + mu,
        q,
        lower,
        upper,
    )


@nb.njit(cache=True)
def nb_logpdf(x, mu, sigma, a, b):
    x = (x - mu) / sigma
    return np.log(
        (1 + x / np.sqrt(a + b + x**2)) ** (a + 0.5)
        * (1 - x / np.sqrt(a + b + x**2)) ** (b + 0.5)
        / (2 ** (a + b - 1) * beta(a, b) * np.sqrt(a + b) * sigma)
    )


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma, a, b):
    return -(nb_logpdf(x, mu, sigma, a, b)).sum()



import warnings
from functools import reduce
from operator import mul

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import tri

from preliz.internal.plot_helper import repr_to_matplotlib
from preliz.internal.special import gammaln


def get_cols_rows(n_plots):
    """Get number of columns and rows for a grid of plots."""
    if n_plots <= 3:
        return 1, n_plots
    else:
        rows = int(np.ceil(n_plots / 3))
        return 3, rows


class DirichletOnSimplex:
    def __init__(self, alpha):
        """Create Dirichlet distribution with parameter `alpha`.

        Adapted from Thomas Boggs' blogpost
        https://blog.bogatron.net/blog/2014/02/02/visualizing-dirichlet-distributions/
        """
        self._alpha = np.array(alpha)
        self._coef = np.exp(
            gammaln(np.sum(self._alpha)) - np.sum([gammaln(a) for a in self._alpha])
        )

        self._corners = np.array([[0.0, 0.0], [1.0, 0.0], [0.5, 0.75**0.5]])
        self._triangle = tri.Triangulation(self._corners[:, 0], self._corners[:, 1])
        self._midpoints = [
            (self._corners[(i + 1) % 3] + self._corners[(i + 2) % 3]) / 2.0 for i in range(3)
        ]

        refiner = tri.UniformTriRefiner(self._triangle)
        self.trimesh = refiner.refine_triangulation(subdiv=8)
        self.pvals = np.nan_to_num(
            [self.pdf(self.xy2bc(xy)) for xy in zip(self.trimesh.x, self.trimesh.y)]
        )
        self.ok = True
        if not np.any(self.pvals):
            self.ok = False
            warnings.warn(
                "The joint pdf is to concentrated to plot, use `marginals=True` instead",
                stacklevel=2,
            )

    def xy2bc(self, x_y, tol=1.0e-3):
        """
        Convert 2D Cartesian coordinates to barycentric coordinates.

        Parameters
        ----------
        xy : A length-2 sequence containing the x and y value.
        """
        bsc = [
            (self._corners[i] - self._midpoints[i]).dot(x_y - self._midpoints[i]) / 0.75
            for i in range(3)
        ]
        return np.clip(bsc, tol, 1.0 - tol)

    def pdf(self, x):
        """Return pdf value for `x`."""
        return self._coef * reduce(mul, [xx ** (aa - 1) for (xx, aa) in zip(x, self._alpha)])

    def plot(self, ax=None):
        """
        Draws pdf contours over the 2-simplex.

        Parameters
        ----------
        dist : A distribution instance with a `pdf` method.
        nlevels: int
            Number of contours to draw.
        subdiv: int
            Number of recursive mesh subdivisions to create.
        """
        hdi_probs = [0.1, 0.5, 0.94]
        contour_levels = find_hdi_contours(self.pvals, hdi_probs)
        if all(contour_levels == contour_levels[0]):
            ax.tricontourf(self.trimesh, self.pvals)
        else:
            ax.tricontour(
                self.trimesh,
                self.pvals,
                levels=contour_levels,
            )
            ax.triplot(self._triangle, color="0.8", linestyle="--", linewidth=2)
        ax.axis("equal")
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 0.75**0.5)
        ax.axis("off")


def plot_dirichlet(
    dist,
    representation,
    marginals,
    pointinterval,
    interval,
    levels,
    support,
    legend,
    figsize,
    axes,
    xy_lim="auto",
):
    """Plot pdf, cdf or ppf of Dirichlet distribution."""
    alpha = dist.alpha
    dim = len(alpha)

    if figsize is None:
        figsize = (12, 4)

    if isinstance(xy_lim, tuple):
        xlim = xy_lim[:2]
        ylim = xy_lim[2:]

    if marginals:
        a_0 = alpha.sum()
        cols, rows = get_cols_rows(dim)

        if axes is None:
            fig, axes = plt.subplots(cols, rows, figsize=figsize, sharex=True, sharey=True)
            axes = axes.flatten()
        if len(axes) > dim:
            for ax in axes[dim:]:
                ax.remove()

        for a_i, ax in zip(alpha, axes):
            marginal_dist = dist.marginal(a_i, a_0 - a_i)
            if xy_lim == "both":
                xlim = marginal_dist._finite_endpoints("full")
                xvals = marginal_dist.xvals("restricted")
                if representation == "pdf":
                    max_pdf = np.max(marginal_dist.pdf(xvals))
                    ylim = (-max_pdf * 0.075, max_pdf * 1.5)
                elif representation == "ppf":
                    max_ppf = marginal_dist.ppf(0.999)
                    ylim = (-max_ppf * 0.075, max_ppf * 1.5)
            if representation == "pdf":
                marginal_dist.plot_pdf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    support=support,
                    legend=False,
                    ax=ax,
                )
            elif representation == "cdf":
                marginal_dist.plot_cdf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    support=support,
                    legend=False,
                    ax=ax,
                )
            elif representation == "ppf":
                marginal_dist.plot_ppf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    legend=False,
                    ax=ax,
                )
            if xy_lim != "auto" and representation != "ppf":
                ax.set_xlim(*xlim)
            if xy_lim != "auto" and representation != "cdf":
                ax.set_ylim(*ylim)

        if legend == "title":
            fig.text(0.5, 1, repr_to_matplotlib(dist), ha="center", va="center")

    elif dim == 3:
        dirichlet_ = DirichletOnSimplex(alpha)
        if dirichlet_.ok:
            if axes is None:
                _, axes = plt.subplots(1, 1)
            dirichlet_.plot(ax=axes)
            if legend == "title":
                axes.set_title(repr_to_matplotlib(dist))
    else:
        raise ValueError("joint only works for Dirichlet of dim=3")

    return axes


def joint_normal(dist, ax):
    """Plot pdf of joint normal distribution."""
    x_max = dist.mu[0] + 3 * dist.cov[0, 0] ** 0.5
    y_max = dist.mu[1] + 3 * dist.cov[1, 1] ** 0.5
    x_min = dist.mu[0] - 3 * dist.cov[0, 0] ** 0.5
    y_min = dist.mu[1] - 3 * dist.cov[1, 1] ** 0.5
    x, y = np.mgrid[x_min:x_max:0.01, y_min:y_max:0.01]

    pos = np.dstack((x, y))

    density = dist.pdf(pos)
    contour_levels = find_hdi_contours(density, [0.1, 0.5, 0.94])
    ax.contour(x, y, density, levels=contour_levels)
    ax.set_aspect("equal")


def find_hdi_contours(density, hdi_probs):
    """
    Find contours enclosing regions of highest posterior density.

    Parameters
    ----------
    density : array-like
    hdi_probs : array-like
        An array of highest density interval confidence probabilities.

    Returns
    -------
    contour_levels : list
        The contour levels corresponding to the given HDI probabilities.
    """
    # Using the algorithm from corner.py
    sorted_density = np.sort(density, axis=None)[::-1]
    csd = sorted_density.cumsum()
    csd /= csd[-1]

    contours = np.empty_like(hdi_probs)
    for idx, hdi_prob in enumerate(hdi_probs):
        try:
            contours[idx] = sorted_density[csd <= hdi_prob][-1]
        except IndexError:
            contours[idx] = sorted_density[0]

    contours.sort()

    return contours


def plot_mvnormal(
    dist,
    representation,
    marginals,
    pointinterval,
    interval,
    levels,
    support,
    legend,
    figsize,
    axes,
    xy_lim="auto",
):
    """Plot pdf, cdf or ppf of Multivariate Normal distribution."""
    mu = dist.mu
    sigma = dist.rv_frozen.var() ** 0.5
    dim = len(mu)

    if figsize is None:
        figsize = (12, 4)

    if isinstance(xy_lim, tuple):
        xlim = xy_lim[:2]
        ylim = xy_lim[2:]

    if marginals:
        cols, rows = get_cols_rows(dim)

        if axes is None:
            fig, axes = plt.subplots(cols, rows, figsize=figsize, sharex=True, sharey=True)
            axes = axes.flatten()
        if len(axes) > dim:
            for ax in axes[dim:]:
                ax.remove()

        for mu_i, sigma_i, ax in zip(mu, sigma, axes):
            marginal_dist = dist.marginal(mu_i, sigma_i)
            if xy_lim == "both":
                xlim = marginal_dist._finite_endpoints("full")
                xvals = marginal_dist.xvals("restricted")
                if representation == "pdf":
                    max_pdf = np.max(marginal_dist.pdf(xvals))
                    ylim = (-max_pdf * 0.075, max_pdf * 1.5)
                elif representation == "ppf":
                    max_ppf = marginal_dist.ppf(0.999)
                    ylim = (-max_ppf * 0.075, max_ppf * 1.5)
            if representation == "pdf":
                marginal_dist.plot_pdf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    support=support,
                    legend=False,
                    ax=ax,
                )
            elif representation == "cdf":
                marginal_dist.plot_cdf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    support=support,
                    legend=False,
                    ax=ax,
                )
            elif representation == "ppf":
                marginal_dist.plot_ppf(
                    pointinterval=pointinterval,
                    interval=interval,
                    levels=levels,
                    legend=False,
                    ax=ax,
                )
            if xy_lim != "auto" and representation != "ppf":
                ax.set_xlim(*xlim)
            if xy_lim != "auto" and representation != "cdf":
                ax.set_ylim(*ylim)
        if legend == "title":
            fig.text(0.5, 1, repr_to_matplotlib(dist), ha="center", va="center")

    elif dim == 2:
        if axes is None:
            _, axes = plt.subplots(1, 1)
        joint_normal(dist, axes)
        if legend == "title":
            axes.set_title(repr_to_matplotlib(dist))
    else:
        raise ValueError("joint only works for Multivariate Normal of dim=2")

    return axes



"""BetaBinomial probability distribution."""

import numba as nb
import numpy as np

from preliz.distributions.distributions import Discrete
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import find_ppf, optimize_ml, optimize_moments
from preliz.internal.special import betaln, cdf_bounds


class BetaBinomial(Discrete):
    R"""
    Beta-binomial distribution.

    Equivalent to binomial random variable with success probability
    drawn from a beta distribution.

    The pmf of this distribution is

    .. math::

       f(x \mid \alpha, \beta, n) =
           \binom{n}{x}
           \frac{B(x + \alpha, n - x + \beta)}{B(\alpha, \beta)}

    .. plot::
        :context: close-figs


        from preliz import BetaBinomial, style
        style.use('preliz-doc')
        alphas = [0.5, 1, 2.3]
        betas = [0.5, 1, 2]
        n = 10
        for a, b in zip(alphas, betas):
            BetaBinomial(a, b, n).plot_pdf()

    ========  =================================================================
    Support   :math:`x \in \{0, 1, \ldots, n\}`
    Mean      :math:`n \dfrac{\alpha}{\alpha + \beta}`
    Variance  :math:`\dfrac{n \alpha \beta (\alpha+\beta+n)}{(\alpha+\beta)^2 (\alpha+\beta+1)}`
    ========  =================================================================

    Parameters
    ----------
    n : int
        Number of Bernoulli trials (n >= 0).
    alpha : float
        alpha > 0.
    beta : float
        beta > 0.
    """

    def __init__(self, alpha=None, beta=None, n=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, beta, n)

    def _parametrization(self, alpha=None, beta=None, n=None):
        self.alpha = alpha
        self.beta = beta
        self.n = n
        self.params = (self.alpha, self.beta, self.n)
        self.param_names = ("alpha", "beta", "n")
        self.params_support = ((eps, np.inf), (eps, np.inf), (eps, np.inf))
        if all_not_none(alpha, beta):
            self._update(alpha, beta, n)

    def _update(self, alpha, beta, n):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.n = np.int64(n)
        self.params = (self.alpha, self.beta, self.n)
        self.support = (0, self.n)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        if isinstance(x, np.ndarray | list | tuple):
            cdf_values = np.zeros_like(x, dtype=float)
            for i, val in enumerate(x):
                x_vals = np.arange(0, val + 1)
                cdf_values[i] = np.sum(self.pdf(x_vals))
            return cdf_bounds(cdf_values, x, *self.support)
        else:
            x_vals = np.arange(0, x + 1)
            return cdf_bounds(np.sum(self.pdf(x_vals)), x, *self.support)

    def ppf(self, q):
        q = np.asarray(q)
        return find_ppf(self, q)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta, self.n, *self.support)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta, self.n, *self.support)

    def entropy(self):
        x_values = self.xvals("full")
        logpdf = self.logpdf(x_values)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        return self.n * self.alpha / (self.alpha + self.beta)

    def mode(self):
        return np.clip(
            np.floor((self.n + 1) * ((self.alpha - 1) / (self.alpha + self.beta - 2))).astype(int),
            0,
            self.n,
        )

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (
            self.n
            * self.alpha
            * self.beta
            * (self.alpha + self.beta + self.n)
            / ((self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1))
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (
            (self.alpha + self.beta + 2 * self.n)
            * (self.beta - self.alpha)
            / (self.alpha + self.beta + 2)
            * (
                (1 + self.alpha + self.beta)
                / (self.n * self.alpha * self.beta * (self.alpha + self.beta + self.n))
            )
            ** 0.5
        )

    def kurtosis(self):
        alpha, beta, n = self.alpha, self.beta, self.n
        alpha_beta_sum = alpha + beta
        alpha_beta_product = alpha * beta
        numerator = ((alpha_beta_sum) ** 2) * (1 + alpha_beta_sum)
        denominator = (
            (n * alpha_beta_product)
            * (alpha_beta_sum + 2)
            * (alpha_beta_sum + 3)
            * (alpha_beta_sum + n)
        )
        left = numerator / denominator
        right = (
            (alpha_beta_sum) * (alpha_beta_sum - 1 + 6 * n)
            + 3 * alpha_beta_product * (n - 2)
            + 6 * n**2
        )
        right -= (3 * alpha_beta_product * n * (6 - n)) / alpha_beta_sum
        right -= (18 * alpha_beta_product * n**2) / (alpha_beta_sum) ** 2
        return (left * right) - 3

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.binomial(
            n=self.n, p=random_state.beta(self.alpha, self.beta, size=size)
        )

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta, n, lower, upper):
    if x < lower:
        return -np.inf
    if x > upper:
        return -np.inf
    else:
        combiln = -np.log(n + 1) - betaln(n - x + 1, x + 1)
        return combiln + betaln(x + alpha, n - x + beta) - betaln(alpha, beta)


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta, n, lower, upper):
    return -(nb_logpdf(x, alpha, beta, n, lower, upper)).sum()
